@@ -1,6 +1,7 @@
 import json
 import os
 import shutil
+import subprocess
 
 import hvac
 import requests
@@ -20,6 +21,10 @@ class RundeckJobs(object):
     wait_random_range: tuple = (60, 180)
     number_of_retries: int = 3
 
+    @retry(wait=wait_random(*wait_random_range), stop=stop_after_attempt(3))
+    def run_command(self, command, cwd, check=True):
+        subprocess.run(command, cwd=cwd, check=check)
+
     def search_uptimerobot(
             self,
             deployment_action: str = os.environ.get(
@@ -116,32 +121,53 @@ class RundeckJobs(object):
         os.environ["TF_IN_AUTOMATION"] = "true"
 
         shutil.copytree(f"{terraform_code_dir}/", f"{config_dir_location}/")
-        zx.run_shell_print(
-            f"cd {config_dir_location} && \
-            terraform init -input=false -reconfigure -force-copy -backend-config {backend_config_file_location}")
+        # zx.run_shell_print(
+        #     f"cd {config_dir_location} && \
+        #     terraform init -input=false -reconfigure -force-copy -backend-config {backend_config_file_location}")
+        self.run_command(["terraform", "init", "-input=false", "-reconfigure", "-force-copy",
+                          f"-backend-config", backend_config_file_location],
+                         cwd=config_dir_location, check=True)
         try:
-            zx.run_shell_print(
-                f"cd {config_dir_location} && \
-                terraform workspace select {fqdn}.{deployment_environment}.{datacenter}")
-        except:
-            zx.run_shell_print(
-                f"cd {config_dir_location} && \
-                terraform workspace new {fqdn}.{deployment_environment}.{datacenter}")
+            # zx.run_shell_print(
+            #     f"cd {config_dir_location} && \
+            #     terraform workspace select {fqdn}.{deployment_environment}.{datacenter}")
+            subprocess.run(["terraform", "workspace", "select", f"{fqdn}.{deployment_environment}.{datacenter}"],
+                           cwd=config_dir_location, check=True)
+        except subprocess.CalledProcessError:
+            # zx.run_shell_print(
+            #     f"cd {config_dir_location} && \
+            #     terraform workspace new {fqdn}.{deployment_environment}.{datacenter}")
+            subprocess.run(["terraform", "workspace", "new", f"{fqdn}.{deployment_environment}.{datacenter}"],
+                           cwd=config_dir_location, check=True)
 
         plan_file_location = f"{config_dir_location}/config.plan"
 
         if deployment_action == "apply":
-            zx.run_shell_print(
-                f"cd {config_dir_location} && \
-                terraform plan -out {plan_file_location} -input=false -var-file {vars_file_location} && \
-                terraform apply -input=false -auto-approve {plan_file_location}")
+            # zx.run_shell_print(
+            #     f"cd {config_dir_location} && \
+            #     terraform plan -out {plan_file_location} -input=false -var-file {vars_file_location} && \
+            #     terraform apply -input=false -auto-approve {plan_file_location}")
+            self.run_command(
+                ["terraform", "plan", "-out", plan_file_location,
+                 "-input=false", "-var-file", vars_file_location],
+                cwd=config_dir_location, check=True)
+            self.run_command(["terraform", "apply", "-input=false", "-auto-approve", plan_file_location],
+                             cwd=config_dir_location, check=True)
         elif deployment_action == "destroy":
-            zx.run_shell_print(
-                f"cd {config_dir_location} && \
-                terraform plan -destroy -out {plan_file_location} -input=false -var-file {vars_file_location} && \
-                terraform apply -input=false -auto-approve {plan_file_location} && \
-                terraform workspace select default && \
-                terraform workspace delete {fqdn}.{deployment_environment}.{datacenter}")
+            # zx.run_shell_print(
+            #     f"cd {config_dir_location} && \
+            #     terraform plan -destroy -out {plan_file_location} -input=false -var-file {vars_file_location} && \
+            #     terraform apply -input=false -auto-approve {plan_file_location} && \
+            #     terraform workspace select default && \
+            #     terraform workspace delete {fqdn}.{deployment_environment}.{datacenter}")
+            self.run_command(["terraform", "plan", "-destroy", "-out", plan_file_location, "-input=false", "-var-file",
+                              vars_file_location], cwd=config_dir_location, check=True)
+            self.run_command(["terraform", "apply", "-input=false", "-auto-approve", plan_file_location],
+                             cwd=config_dir_location, check=True)
+            self.run_command(["terraform", "workspace", "select",
+                              "default"], cwd=config_dir_location, check=True)
+            self.run_command(["terraform", "workspace", "delete", f"{fqdn}.{deployment_environment}.{datacenter}"],
+                             cwd=config_dir_location, check=True)
 
     @retry(wait=wait_random(*wait_random_range), stop=stop_after_attempt(number_of_retries))
     def hcloud_web_solutions(
@@ -240,11 +266,14 @@ class RundeckJobs(object):
         os.environ["TF_IN_AUTOMATION"] = "true"
 
         shutil.copytree(f"{terraform_code_dir}/", f"{config_dir_location}/")
-        zx.run_shell_print(
-            f"cd {config_dir_location} && \
-            terraform init -input=false -reconfigure -force-copy \
-            -backend-config {backend_config_file_location}"
-        )
+        # zx.run_shell_print(
+        #     f"cd {config_dir_location} && \
+        #     terraform init -input=false -reconfigure -force-copy \
+        #     -backend-config {backend_config_file_location}"
+        # )
+        self.run_command(["terraform", "init", "-input=false", "-reconfigure", "-force-copy", "-backend-config",
+                          backend_config_file_location], cwd=config_dir_location,
+                         check=True)
 
         vars_file_location = f"{config_dir_location}/{fqdn}.auto.tfvars"
         vars_file_content = f"""fqdn = "{fqdn}"
@@ -278,45 +307,85 @@ class RundeckJobs(object):
             f.write(vars_file_content)
 
         if deployment_action == "apply":
-            zx.run_shell_print(f"cd {config_dir_location} && \
-            terraform apply -input=false -auto-approve")
+            # zx.run_shell_print(f"cd {config_dir_location} && \
+            # terraform apply -input=false -auto-approve")
+            self.run_command(["terraform", "apply", "-input=false", "-auto-approve"], cwd=config_dir_location,
+                             check=True)
             if deployment_environment == "testing":
                 deployment_action = "destroy"
                 destroy_delay = "3d"
-                zx.run_shell_print(
-                    f"rd run --project main --job Datacenter/Hetzner/Cloud/Deploy --delay {destroy_delay} -- \
-                     -deployment_action {deployment_action} \
-                     -fqdn {fqdn} \
-                     -project_name {project_name} \
-                     -sld {sld} \
-                     -webserver_netdata_fqdn {webserver_netdata_fqdn} \
-                     -mariadb_server_netdata_fqdn {mariadb_server_netdata_fqdn} \
-                     -webserver_fqdn {webserver_fqdn} \
-                     -mariadb_server_fqdn {mariadb_server_fqdn} \
-                     -deployment_environment {deployment_environment} \
-                     -setup_wordpress {setup_wordpress} \
-                     -wordpress_site_title {wordpress_site_title} \
-                     -with_wordpress_lifter_lms {with_wordpress_lifter_lms} \
-                     -wordpress_lms_config_repo {wordpress_lms_config_repo} \
-                     -wordpress_lms_config_repo_script_dir {wordpress_lms_config_repo_script_dir} \
-                     -wordpress_lms_config_repo_script_name {wordpress_lms_config_repo_script_name} \
-                     -with_lifter_lms_loadtest_course {with_lifter_lms_loadtest_course} \
-                     -with_internal_mariadb {with_internal_mariadb} \
-                     -hcloud_token {hcloud_token} \
-                     -vault_login_username {vault_login_username} \
-                     -vault_login_password {vault_login_password} \
-                     -atlas_mongo_public_key {atlas_mongo_public_key} \
-                     -atlas_mongo_private_key {atlas_mongo_private_key} \
-                     -web_server_image {web_server_image} \
-                     -mariadb_server_image {mariadb_server_image} \
-                     -web_server_type {web_server_type} \
-                     -mariadb_server_type {mariadb_server_type} \
-                     -jira_issue_key {jira_issue_key} \
-                     -tag_names {os.environ.get('RD_OPTION_TAG_NAMES')} \
-                     -auto_apply {auto_apply} \
-                     -execution_mode {execution_mode}"
+                # zx.run_shell_print(
+                #     f"rd run --project main --job Datacenter/Hetzner/Cloud/Deploy --delay {destroy_delay} -- \
+                #      -deployment_action {deployment_action} \
+                #      -fqdn {fqdn} \
+                #      -project_name {project_name} \
+                #      -sld {sld} \
+                #      -webserver_netdata_fqdn {webserver_netdata_fqdn} \
+                #      -mariadb_server_netdata_fqdn {mariadb_server_netdata_fqdn} \
+                #      -webserver_fqdn {webserver_fqdn} \
+                #      -mariadb_server_fqdn {mariadb_server_fqdn} \
+                #      -deployment_environment {deployment_environment} \
+                #      -setup_wordpress {setup_wordpress} \
+                #      -wordpress_site_title {wordpress_site_title} \
+                #      -with_wordpress_lifter_lms {with_wordpress_lifter_lms} \
+                #      -wordpress_lms_config_repo {wordpress_lms_config_repo} \
+                #      -wordpress_lms_config_repo_script_dir {wordpress_lms_config_repo_script_dir} \
+                #      -wordpress_lms_config_repo_script_name {wordpress_lms_config_repo_script_name} \
+                #      -with_lifter_lms_loadtest_course {with_lifter_lms_loadtest_course} \
+                #      -with_internal_mariadb {with_internal_mariadb} \
+                #      -hcloud_token {hcloud_token} \
+                #      -vault_login_username {vault_login_username} \
+                #      -vault_login_password {vault_login_password} \
+                #      -atlas_mongo_public_key {atlas_mongo_public_key} \
+                #      -atlas_mongo_private_key {atlas_mongo_private_key} \
+                #      -web_server_image {web_server_image} \
+                #      -mariadb_server_image {mariadb_server_image} \
+                #      -web_server_type {web_server_type} \
+                #      -mariadb_server_type {mariadb_server_type} \
+                #      -jira_issue_key {jira_issue_key} \
+                #      -tag_names {os.environ.get('RD_OPTION_TAG_NAMES')} \
+                #      -auto_apply {auto_apply} \
+                #      -execution_mode {execution_mode}"
+                # )
+                self.run_command(
+                    ["rd", "run", "--project", "main", "--job", "Datacenter/Hetzner/Cloud/Deploy", "--delay",
+                     destroy_delay, "--",
+                     "-deployment_action", f"{deployment_action}",
+                     "-fqdn", f"{fqdn}",
+                     "-project_name", f"{project_name}",
+                     "-sld", f"{sld}",
+                     "-webserver_netdata_fqdn", f"{webserver_netdata_fqdn}",
+                     "-mariadb_server_netdata_fqdn", f"{mariadb_server_netdata_fqdn}",
+                     "-webserver_fqdn", f"{webserver_fqdn}",
+                     "-mariadb_server_fqdn", f"{mariadb_server_fqdn}",
+                     "-deployment_environment", f"{deployment_environment}",
+                     "-setup_wordpress", f"{setup_wordpress}",
+                     "-wordpress_site_title", f"{wordpress_site_title}",
+                     "-with_wordpress_lifter_lms", f"{with_wordpress_lifter_lms}",
+                     "-wordpress_lms_config_repo", f"{wordpress_lms_config_repo}",
+                     "-wordpress_lms_config_repo_script_dir", f"{wordpress_lms_config_repo_script_dir}",
+                     "-wordpress_lms_config_repo_script_name", f"{wordpress_lms_config_repo_script_name}",
+                     "-with_lifter_lms_loadtest_course", f"{with_lifter_lms_loadtest_course}",
+                     "-with_internal_mariadb", f"{with_internal_mariadb}",
+                     "-hcloud_token", f"{hcloud_token}",
+                     "-vault_login_username", f"{vault_login_username}",
+                     "-vault_login_password", f"{vault_login_password}",
+                     "-atlas_mongo_public_key", f"{atlas_mongo_public_key}",
+                     "-atlas_mongo_private_key", f"{atlas_mongo_private_key}",
+                     "-web_server_image", f"{web_server_image}",
+                     "-mariadb_server_image", f"{mariadb_server_image}",
+                     "-web_server_type", f"{web_server_type}",
+                     "-mariadb_server_type", f"{mariadb_server_type}",
+                     "-jira_issue_key", f"{jira_issue_key}",
+                     "-tag_names", f"{os.environ.get('RD_OPTION_TAG_NAMES')}",
+                     "-auto_apply", f"{auto_apply}",
+                     "-execution_mode", f"{execution_mode}"],
                 )
 
         elif deployment_action == "destroy":
-            zx.run_shell_print(f"cd {config_dir_location} && \
-            terraform destroy -input=false -auto-approve")
+            # zx.run_shell_print(f"cd {config_dir_location} && \
+            # terraform destroy -input=false -auto-approve")
+            self.run_command(["terraform", "destroy", "-input=false", "-auto-approve"], cwd=config_dir_location,
+                             shell=True)
+            self.terraform_cloud_workspace(deployment_action, fqdn, datacenter, organization, terraform_cloud_token,
+                                           auto_apply, execution_mode, deployment_environment, terraform_code_dir)
